var Parse = function(opts = null) {
	var defaultOpts = {};
	this.opts = (opts === null) ? defaultOpts : opts;

	this.on = function(l, u, r, i) {
		return function(n, f) {
			r = l[n] = l[n] || [], i = -1;
			if (f && f.call) r.push(f);
			else while (r[++i]) r[i].apply(u, arguments);
		}
	}({});

	var log = this.log = function() {
		var content = '';
		for(var i=0;i<arguments.length;i++) {
			content += arguments[i] + ' ';
		}
		console.log('%c[parse]%c ' + content, 'color: #673AB7;font-weight:bold;', 'color: #666;');
	}

	this.initialize = function() {
		this.clients = {};
		this.rooms = {};
		this.online = 0;
		this.currentRoom = null;
		this.url = null;
	}

	this.connect = function(url, query) {
		this.url = url;
		this.socket = io.connect(url, { query });
		this.setupListeners();
		return this.socket;
	}

	this.setupListeners = function() {
		this.socket.on('connect', () => {
			if(this.opts.loglevel === 'debug') log('Connected to '+this.url+' as ' + this.socket.id);
			this.on('connect', socket);
		});

		this.socket.on('get_server_info', serverInfo => {
			if(this.opts.loglevel === 'debug') log('get_server_info');
			this.online = serverInfo.online;
			this.clients = serverInfo.clients;
			this.rooms = serverInfo.rooms;
			this.updateStats();
			this.on('get_server_info', serverInfo);
		});

		this.socket.on('get_room_info', roomInfo => {
			if(this.opts.loglevel === 'debug') log('get_room_info');
			if(this.opts.loglevel === 'debug') log(roomInfo);
			this.on('get_room_info', roomInfo);
		});

		this.socket.on('client_connect', client => {
			if(this.opts.loglevel === 'debug') log('client_connect');
			if(!this.clients.containsKey(client.id)) {
				this.clients[client.id] = client;
				this.updateStats();
			}
			this.on('client_connect', client);
		});

		this.socket.on('client_disconnect', client => {
			if(this.opts.loglevel === 'debug') log('client_disconnect');
			if(this.clients.containsKey(client.id)) {
				delete this.clients[client.id];
				this.updateStats();
			}
			this.on('client_disconnect', client);
		});

		this.socket.on('create_room', room => {
			if(this.opts.loglevel === 'debug') log('create_room');
			if(!this.rooms.containsKey(room.id)) {
				this.rooms[room.id] = room;
				if(this.opts.loglevel === 'debug') log('Room created: ' + room.name);
				this.updateStats();
			}
			this.on('create_room', room);
		});

		this.socket.on('delete_room', room => {
			if(this.opts.loglevel === 'debug') log('delete_room');
			if(this.rooms.containsKey(room.id)) {
				var room = this.rooms[room.id];
				delete this.rooms[room.id];
				if(this.opts.loglevel === 'debug') log('Room deleted: ' + room.name);
				this.updateStats();
			}
			this.on('delete_room', room);
		});

		this.socket.on('join_room', data => {
			if(this.opts.loglevel === 'debug') log('join_room');
			this.updateRoom(data);
			this.on('join_room', data);
		});

		this.socket.on('leave_room', data => {
			if(this.opts.loglevel === 'debug') log('leave_room');
			this.updateRoom(data);
			this.on('leave_room', data);
		});

		this.socket.on('room_connect', data => {
			if(this.opts.loglevel === 'debug') log('room_connect');
			var connectingClient = data.client;
			this.updateRoom(data.room);
			this.on('room_connect', data);
		});

		this.socket.on('room_disconnect', data => {
			if(this.opts.loglevel === 'debug') log('room_disconnect');
			var disconnectingClient = data.client;
			this.updateRoom(data.room);
			this.on('room_disconnect', data);
		});

		this.socket.on('receive_packet', packetData => {
			if(this.opts.loglevel === 'debug') log('receive_packet');
			var data = JSON.parse(packetData);
			this.on('receive_packet', data);
		});
	}

	this.updateStats = function() {
		document.title = 'Parse | online: ' + this.clients.length() + ' | rooms: ' + this.rooms.length();
	}

	this.updateRoom = function(room) {
		var roomClients = filterSelfFromClients(room.clients, socket.id);
		room.clients = roomClients;
		if(this.rooms.containsKey(room.id)) {
			var clientRoom = this.rooms[room.id];
			clientRoom = room;
		}
	}

	this.createRoom = function(name) { socket.emit('create_room', { name, creator: 'tester' }); }
	this.deleteRoom = function(id) { socket.emit('delete_room', id); }
	this.joinRoom = function(id) { socket.emit('join_room', id); }
	this.leaveRoom = function(id) { socket.emit('leave_room', id);	}
	this.joinRoomByName = function(name) { rooms.forEach(room => { if(room.name === name) { socket.emit('join_room', room.id); } }); }
	this.leaveRoomByName = function(name) { rooms.forEach(room => { if(room.name === name) { socket.emit('leave_room', room.id); } }); }
	this.sendPacket = function(packet, inRoom) { socket.emit('send_packet', { sender: 'tester', packet, inRoom, timestamp: (new Date()).toString() }); }

	this.filterSelfFromClients = function(obj, id) {
		var newObj = {};
		obj.forEach(el => { if(el.id !== id) newObj[el.id] = el; });
		return newObj;
	}

	this.initialize();
	return this;
}

